#include "flow_fun.h"
#include "flow_pub.h"
#include "af_capture.h"

#define MAXFD 64
#define PATH_SEP '/'
void open_log();
void (*write_log_ptr)(char *fmt, ...);
void write_log_enable(char *fmt, ...);
void write_log_disable(char *fmt, ...);
void close_log();
void create_cap_socket();
void create_send_socket();
void change_send_sockaddr(char *paddr,UPORT port);
void daemon_init(const char *pname,int facility);
void check_runuser();
void check_lockpid();
void write_newpid();
void nfc_exited();
void nfc_killed(int signo);
static void init_program();
static void printcmdline(char *pname);
static void handleopt(int argc, char* argv[]);

static char *prog_name;
extern int daemon_proc;
static int daemonflag=0;  /*1 daemon  0 userprogram*/
static unsigned int send_port=SEND_PORT;
static char send_paddr[16]=SEND_PADDR;
static char logfile[512]=LOGFILE;
static int logenable=LOGENABLE;

int main (int argc, char *argv[]) 
{
	pthread_t read_data_thr;
	pthread_t process_data_thr;
	pthread_t send_data_thr;
	sigset_t oldmask,newmask;
	
	if ((prog_name=strrchr(argv[0],PATH_SEP))!=NULL)
		prog_name++;
	else
		prog_name=argv[0];
		
	handleopt(argc,argv);
	check_runuser();
	check_lockpid();
	if (daemonflag==1) daemon_init(prog_name,LOG_LOCAL3);
	write_newpid();
	atexit(nfc_exited);
	init_program();
	
	sigemptyset(&newmask);
	sigaddset(&newmask,SIGTERM);
	sigaddset(&newmask,SIGINT);
	sigaddset(&newmask,SIGQUIT);
	sigaddset(&newmask,SIGILL);
	if (sigprocmask(SIG_BLOCK,&newmask,&oldmask)<0) {
		write_log("SIG_BLOCK error : %s",strerror(errno));
		exit(EXIT_ERROR);
	}
	Signal(SIGTERM,nfc_killed);
	Signal(SIGINT,nfc_killed);
	Signal(SIGQUIT,nfc_killed);
	Signal(SIGILL,nfc_killed);
	
	write_log("linuxflow main thread tid = %d   pid = %d",pthread_self(),getpid());
	
	Pthread_create(&send_data_thr,NULL,send_data_thread,NULL);
	usleep(10000); puts("\t send thread");
	
	Pthread_create(&process_data_thr,NULL,process_data_thread,NULL);
	usleep(10000); puts("\t process thread");
	
	Pthread_create(&read_data_thr,NULL,read_data_thread,NULL);
	puts("\t read thread");

	/*** maybe add code for restart and reread configure files ***/
	/*** and output logs and debug information		   ***/
	while (1) {
		sigsuspend(&oldmask);
		write_log("received a signal");
	}
	exit(0);
}

/*
 * open log file to write
 */
void open_log() 
{
	if ((log_fp=fopen(logfile,"a"))==NULL) {
		printf("%s error in open logfile!\n",prog_name);
		exit(EXIT_ERROR);
	} 
	setvbuf(log_fp,NULL,_IOLBF,0);
	
}
/*
 * write log file 
 */
void write_log_enable(char *fmt, ...)
{
	va_list argp;
	Pthread_mutex_lock(&log_fp_mutex);
	//may change add thread name
	fprintf(log_fp,"%s : ",prog_name);
	#ifdef DEBUG
	fprintf(stderr,"%s : ",prog_name);
	#endif
	va_start(argp,fmt);
	vfprintf(log_fp,fmt,argp);
	#ifdef DEBUG
	vfprintf(stderr,fmt,argp);
	#endif
	va_end(argp);
	fprintf(log_fp,"\n");
	#ifdef DEBUG
	fprintf(stderr,"\n");
	#endif
	Pthread_mutex_unlock(&log_fp_mutex);
}

void write_log_disable(char *fmt, ...)
{
}

/*
 * close log file
 */
void close_log() 
{
	fclose(log_fp);
}
/* 
 *open AF_CAPTURE socket 
 */
void create_cap_socket() 
{
	int sock;

	sock=socket(AF_CAPTURE,CAP_COPY_FLOW,htons(ETH_P_IP));
	if (sock<0) {
		write_log("main thread : Fatal Error in open AF_CAPTURE socket: %s", strerror(errno));
		exit(EXIT_ERROR);
	}
	cap_sock=sock;
}
/* 
 *open a UDP socket to send out the session data
 */
void create_send_socket()
{
	int sock;
	
	sock=socket(AF_INET,SOCK_DGRAM,0);
	if (sock<0) {
		write_log("main thread : Fatal Error in open send UDP socket: %s",strerror(errno));
		exit(EXIT_ERROR);
	}
	bzero(&send_addr,sizeof(send_addr));
	send_addr.sin_family=AF_INET;
	send_addr.sin_port=htons(send_port);
	send_addr.sin_addr.s_addr=inet_addr(send_paddr);
	send_sock=sock;
}
/*
 * can change the send out place
 */
void change_send_sockaddr(char *paddr,UPORT port)
{
	send_addr.sin_port=htons(port);
	send_addr.sin_addr.s_addr=inet_addr(paddr);
}
/* 
 *initialize the program
 */
static void init_program() 
{
	time_t ticks;
	char strtime[100];
	
	Pthread_mutex_init(&cap_sock_mutex,NULL);
	Pthread_mutex_init(&send_sock_mutex,NULL);
	Pthread_mutex_init(&log_fp_mutex,NULL);
	
	if (logenable) open_log();
	if (logenable) 
		write_log_ptr=write_log_enable;
	else
		write_log_ptr=write_log_disable;
		
	write_log("**************************************************");
	write_log("%s program begin logging",prog_name);
	time(&ticks); strcpy(strtime,ctime(&ticks));
	strtime[strlen(strtime)-1]='\0';
	write_log("Begin time: %s",strtime);
	write_log("program pid = %d",getpid());
	write_log("**************************************************");
	create_cap_socket();
	create_send_socket();
	
}

void daemon_init(const char *pname,int facility)
{
	int i;
	pid_t pid;
	int fileDesc=-1;
	if ((pid=fork())!=0) exit(0);
	
	setsid();
	Signal(SIGHUP,SIG_IGN);
	if ((pid=fork())!=0) exit(0);
	daemon_proc=1;
	umask(0);
	for (i=0;i<MAXFD;i++) {
		close(i);
	}
	fileDesc = open("/dev/null", O_RDWR);/* stdin */
    	(void) dup(fileDesc);	/* stdout */
    	(void) dup(fileDesc);	/* stderr */
	openlog(pname,LOG_PID,facility);
}

void check_runuser()
{
	if (geteuid()) {
		printf("%s only can run by root!\n",prog_name);
		exit(EXIT_ERROR);
	}
}

void check_lockpid() 
{
	pid_t old_pid;
	FILE *fp;
	fp=fopen(PIDFILE,"r");
	if (fp==NULL) return;
	if (1 !=fscanf(fp,"%d",&old_pid)) {
		unlink(PIDFILE);
		return;
	}
	if (-1 == kill(old_pid,0)) {
		unlink(PIDFILE);
		return;
	}
	printf("%s already running as pid %d\n",prog_name,old_pid);
	exit(0);
}

void write_newpid() 
{
	FILE *fp;
	fp=fopen(PIDFILE,"w");
	fprintf(fp,"%d\n",getpid());
	fclose(fp);
}
void nfc_exited()
{
	unlink(PIDFILE);	
}
void nfc_killed(int signo)
{	
	time_t ticks;
	
	write_log("%d Program %s terminated by signal %d",getpid(),prog_name,signo);
	time(&ticks);
	write_log("%d Ended time is %s",getpid(),ctime(&ticks)); 
	exit(0);
}

static void printcmdline(char *pname)
{
	printf("Usage : %s -D -H <ipaddress> -P <port> -L <fullpathlogfilename>\n",pname);
}

static void handleopt(int argc, char* argv[])
{
	int c;
	while (1) {
		c=getopt(argc,argv,"DH:P:L:");
		if (c==-1) break;
		switch (c) {
			case 'D':
				daemonflag=1;
				break;
			case 'H':
				if(strcmp(optarg,"")==0) {
					printf("Missing IPaddress -H should be followed by IPaddress!\n");
					printcmdline(prog_name);
					exit(EXIT_ERROR);
				}
				strcpy(send_paddr,optarg);
				break;
			case 'P':
				send_port=atoi(optarg);
				if(send_port<=0) {
					printf("Missing Port -P should be followed by Port!\n");
					printcmdline(prog_name);
					exit(EXIT_ERROR);
				}
				
				break;
			case 'L':
				if(strcmp(optarg,"")==0) {
					printf("Missing FullPathLogFileName -L should be followed by Logfilename!\n");
					printcmdline(prog_name);
					exit(EXIT_ERROR);
				}
				strcpy(logfile,optarg);
				logenable=1;
				break;
			default:
				printf("Parameter error!\n");
				printcmdline(prog_name);
				exit(EXIT_ERROR);
				break;
		}
	}
	if (optind!=argc) {
		printf("Parameter error!\n");
		printcmdline(prog_name);
		exit(EXIT_ERROR);
	}
}

		
					
		
