#include "flow_fun.h"
#include "flow_pub.h"
#include "ipasct.h"

static HashTable *curr_hash;
static HashTable *last_hash;
static struct list_head	*curr_list;
static struct list_head *last_list;
static struct list_head *free_list;

static time_t next_flush;
static int    free_count;
static pthread_mutex_t free_mutex;
static unsigned long long getpacket=0;
static unsigned long long oldpacket=0;
static unsigned long long getbytes=0;
static unsigned long long oldbytes=0;
static void init_process() 
{
	max_free_item=MAX_FREE_COUNT;
	free_count=0;
	Pthread_mutex_init(&free_mutex,NULL);
	next_flush=(time(NULL)+FLUSH_INTERVAL);
	
	curr_hash=(HashTable *)malloc(sizeof(HashTable));
	if (curr_hash==NULL) {
		write_log("process thread : Error in malloc HashTable curr_hash: %s",strerror(errno));
		exit(EXIT_ERROR);
	}
	InitHashTable(curr_hash,sizeof(struct hash_key)/sizeof(int));
	last_hash=(HashTable *)malloc(sizeof(HashTable));
	if (last_hash==NULL) {
		write_log("process thread : Error in malloc HashTable last_hash: %s",strerror(errno));
		exit(EXIT_ERROR);
	}
	InitHashTable(last_hash,sizeof(struct hash_key)/sizeof(int));
	
	curr_list=(struct list_head *)malloc(sizeof(struct list_head));
	if (curr_list==NULL) {
		write_log("process thread : Error in malloc list structure curr_list: %s",strerror(errno));
		exit(EXIT_ERROR);
	}
	INIT_LIST_HEAD(curr_list);
	last_list=(struct list_head *)malloc(sizeof(struct list_head));
	if (last_list==NULL) {
		write_log("process thread : Error in malloc list structure last_list: %s",strerror(errno));
		exit(EXIT_ERROR);
	}
	INIT_LIST_HEAD(last_list);
	free_list=(struct list_head *)malloc(sizeof(struct list_head));
	if (free_list==NULL) {
		write_log("process thread : Error in malloc list structure free_list: %s",strerror(errno));
		exit(EXIT_ERROR);
	}
	INIT_LIST_HEAD(free_list);
}

static __inline__ int need_reverse(struct flow_type *flowp) {
	return ((flowp->sport<flowp->dport)||((flowp->sport==flowp->dport)&&(flowp->saddr<flowp->daddr)))?1:0;
}

static int copy_key(struct hash_key *key,struct flow_type *flowp)
{
	if(need_reverse(flowp)) {
		key->saddr=flowp->daddr;
		key->daddr=flowp->saddr;
		key->sport=flowp->dport;
		key->dport=flowp->sport;
		key->protocol=flowp->protocol;
		return 1;
	} else {
		key->saddr=flowp->saddr;
		key->daddr=flowp->daddr;
		key->sport=flowp->sport;
		key->dport=flowp->dport;
		key->protocol=flowp->protocol;
		return 0;	
	}
}

struct session_queue *get_item() 
{
	int i;
	struct list_head *list;
	struct session_queue *item;
	struct list_head tmp_list;
	int tmp_count;
	
	Pthread_mutex_lock(&free_mutex);
	if(list_empty(free_list)) {
		Pthread_mutex_unlock(&free_mutex);
		INIT_LIST_HEAD(&tmp_list);
		tmp_count=0;
		for (i=0;i<ALLOC_FREENODE_ONETIME;i++) {
			item=(struct session_queue *)malloc(sizeof(struct session_queue));
			if (item==NULL) {
				write_log("process thread : Error in malloc freenode:%s",strerror(errno));
				exit(EXIT_ERROR);
			}
			list_add(&item->list,&tmp_list);
			tmp_count++;
		}
		Pthread_mutex_lock(&free_mutex);
		list_splice(&tmp_list,free_list);
		free_count+=tmp_count;
	}
	item=NULL;
	if(!list_empty(free_list)) {
		list=free_list->next;
		list_del(list);
		free_count--;
		item=(struct session_queue *)list;
	}
	Pthread_mutex_unlock(&free_mutex);
	return item;
}

void put_item (struct session_queue *item) 
{
	Pthread_mutex_lock(&free_mutex);
	if (free_count>MAX_FREE_COUNT) {
		free(item);
	} else {
		list_add(&item->list,free_list);
		free_count++;
	}
	Pthread_mutex_unlock(&free_mutex);
}

void put_list (struct list_head *send_list,unsigned long list_len) 
{
	int delnum,num;
	struct list_head *next;
	struct list_head *tmp_list;
	max_free_item=max_free_item*FREE_COUNT_RATIO+list_len*(1-FREE_COUNT_RATIO);
	Pthread_mutex_lock(&free_mutex);
	if (free_count+list_len<=max_free_item) {
		list_splice(send_list,free_list);
		free_count+=list_len;
	} else {
		delnum=free_count+list_len-max_free_item;
		Pthread_mutex_unlock(&free_mutex);
		for (num=0,next=send_list->next;num<delnum;num++) {
			tmp_list=next;
			next=next->next;
			list_del(tmp_list);
			free((struct session_queue *)tmp_list);
		}
		Pthread_mutex_lock(&free_mutex);
		list_splice(send_list,free_list);
		free_count+=list_len-delnum;
	}
	Pthread_mutex_unlock(&free_mutex);
	INIT_LIST_HEAD(send_list);
}

static int process(struct flow_type *flowp)
{
	volatile int reversed,newone;
	struct hash_key key;
	struct session_queue *item;
	HashEntry *entry;
	
	reversed=copy_key(&key,flowp);
	entry=CreateHashEntry(curr_hash,(char*)&key,&newone);
	if (newone) {
		item=get_item();
		memcpy(&item->session.key,&key,sizeof(key));
		item->session.flow[0]=item->session.flow[1]=0;
		item->session.flow[reversed]=flowp->length;
		item->session.time[0]=item->session.time[1]=flowp->time;
		SetHashValue(entry,item);
	} else {
		item=(struct session_queue *)GetHashValue(entry);
		if (!item->session.flow[reversed]) newone=1;
		item->session.flow[reversed]+=flowp->length;
		item->session.time[reversed]=flowp->time;
		list_del(&item->list);
	}
	list_add(&item->list,curr_list);
	return newone;
}

static int flush_data() 
{
	HashTable *hash;
	struct list_head *list;
	
	if (send_hash(curr_hash,curr_list)) 
		return -1;
	
	hash=curr_hash;
	curr_hash=last_hash;
	last_hash=hash;
	
	list=curr_list;
	curr_list=last_list;
	last_list=list;
	return 0;
}
		
		
	
	

	
void *process_data_thread (void *dummy) 
{
	char *data;
	int len;
	struct flow_type *flowp;
	
	write_log("linuxflow process thread tid = %d   pid = %d",pthread_self(),getpid());
	init_process();
	
	while(1) {
		data=getdata(&len);
		for (flowp=(struct flow_type *)data;(char*)flowp<data+len;flowp++) {
			process(flowp); getpacket++;
			getbytes+=flowp->length;
		}
		putdata();
		
		if (time(NULL)>=next_flush) {
			if(!flush_data()) {
				while (next_flush<=time(NULL)) next_flush+=FLUSH_INTERVAL;
				//next_flush+=FLUSH_INTERVAL;
				write_log("process thread : processed packet head %ld",getpacket-oldpacket);
				oldpacket=getpacket;
				write_log("process thread : processed bytes %lld",getbytes-oldbytes);
				oldbytes=getbytes;
			}
			
		}
	}
}

		
	
