#include "flow_fun.h"
#include "flow_pub.h"
#include "ipasct.h"


static HashTable *hashptr;
static struct list_head *listptr;
static struct list_head send_head;
static atomic_t	hash_active;
static atomic_t list_active;
static pthread_cond_t active_cond;
static pthread_cond_t processhash_cond;
static pthread_mutex_t send_mutex;

static struct session_type *send_buf;
static struct session_type *curr_send_buf;
static struct session_type *top_send_buf;

static void init_send_data()
{
	int send_buf_max_len;
	
	atomic_set(&hash_active,0);
	atomic_set(&list_active,0);
	pthread_cond_init(&active_cond,NULL);
	pthread_cond_init(&processhash_cond,NULL);
	Pthread_mutex_init(&send_mutex,NULL);
	INIT_LIST_HEAD(&send_head);
	
	send_buf_max_len=SEND_BUF_SIZE/sizeof(struct session_type);
	send_buf=(struct session_type *)malloc(SEND_BUF_SIZE);
	if (send_buf==NULL) {
		write_log("Error in alloc buffer for send data");
		exit(EXIT_ERROR);
	}
	curr_send_buf=send_buf;
	top_send_buf=send_buf+send_buf_max_len;
}

int send_hash (HashTable *hash,struct list_head *list)
{	
	int ret;
	Pthread_mutex_lock(&send_mutex);
	if(atomic_read(&hash_active)==0) {
		hashptr=hash;
		listptr=list;
		atomic_set(&hash_active,1);
		Pthread_cond_signal(&active_cond);
		ret=0;
	} else {
		while (atomic_read(&hash_active)>0) {
			Pthread_cond_wait(&processhash_cond,&send_mutex);
		}
		
		ret=-1;
	}
	Pthread_mutex_unlock(&send_mutex);
	return ret;
}

static void send_data ()
{
	struct session_queue *item;
	struct list_head *next;
	struct list_head *tmp_list;
	struct session_type *record;
	unsigned long send_list_len=0;
	char *hstat;
	if (atomic_read(&hash_active)>0) {
		hstat=NULL;
		#ifdef HASHSTAT
		hstat=HashStats(hashptr);
		write_log("send thread : Hash Table Stats");
		write_log("%s",hstat);
		free(hstat);
		#endif
		//DeleteHashTable(hashptr);
		//InitHashTable(hashptr,sizeof(struct hash_key)/sizeof(int));
		DelInitHashTable(hashptr,sizeof(struct hash_key)/sizeof(int));
		list_add(&send_head,listptr);
		list_del_init(listptr);
		hashptr=NULL;
		listptr=NULL;
		atomic_set(&list_active,1);
		Pthread_mutex_lock(&send_mutex);
		atomic_set(&hash_active,0);
		Pthread_cond_signal(&processhash_cond);
		Pthread_mutex_unlock(&send_mutex);
	}
	if (atomic_read(&list_active)>0) {
		send_list_len=0;
		for (next=send_head.next;next!=&send_head;) {
			if (curr_send_buf >=top_send_buf) {
				//write_log("send thread : send out a buffer");
				sendto(send_sock,(char*)send_buf,SEND_BUF_SIZE,0,(struct sockaddr *)&send_addr,sizeof(send_addr));
				curr_send_buf=send_buf;
			}
			record=curr_send_buf;
			curr_send_buf++;
			tmp_list=next;
			next=next->next;
			//list_del(tmp_list);
			item=(struct session_queue *)tmp_list;
			record->key.saddr=(item->session.key.saddr);
			record->key.daddr=(item->session.key.daddr);
			record->key.sport=htons(item->session.key.sport);
			record->key.dport=htons(item->session.key.dport);
			record->key.protocol=htons(item->session.key.protocol);
			record->time[0]=htonl(item->session.time[0]);
			record->time[1]=htonl(item->session.time[1]);
			record->flow[0]=htonl(item->session.flow[0]);
			record->flow[1]=htonl(item->session.flow[1]);
			//put_item(item);
			send_list_len++;
		}
		write_log("send thread : sended session entrys %ld",send_list_len);
		put_list(&send_head,send_list_len);
		atomic_set(&list_active,0);
	}
}

	





void *send_data_thread(void *dummy) 
{
	write_log("linuxflow send thread tid = %d   pid = %d",pthread_self(),getpid());
	init_send_data();
	
	while(1) {
		Pthread_mutex_lock(&send_mutex);
		while (atomic_read(&hash_active)==0)
			Pthread_cond_wait(&active_cond,&send_mutex);
		Pthread_mutex_unlock(&send_mutex);
		send_data();
	}
}

		
	
