#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif
#include <linux/config.h>
#include <linux/types.h>
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/module.h>
#include <linux/netdevice.h>
#include <linux/skbuff.h>
#include <linux/socket.h>
#include <linux/timer.h>
#include <linux/errno.h>
#include <linux/spinlock.h>
#include <asm/atomic.h>
#include <net/protocol.h>
#include <net/sock.h>
#include "cap_pub.h"
#include "cap_pack.h"
#include "af_capture.h" 
/*
#include <net/ip.h>
#include <asm/system.h>
#include <linux/in.h>
#include <linux/inet.h>
#include <linux/if_packet.h>
#include <linux/sched.h>
#include <linux/fcntl.h>
#include <linux/wireless.h>
#include <linux/kmod.h>
#include <linux/ip.h>
#include <linux/udp.h>
#include <asm/uaccess.h>


#include "debug.h" 
*/

#if defined(MODULE_LICENSE)
        MODULE_LICENSE("GPL");
#endif

struct cap_copy_type * cap_copy_types[MAX_CAPTYPE];
struct proto_ops capture_ops;

/* Use double buffer to server data exchange 
 * Any change to store_slot,hold_slot,free_slot,store_len and hold_len 
 * must be under spinlock (cap->lock).
*/

static int cap_alloc_buffer(struct capture_struct *cap)
{
	if(cap->buffer_size <= 0) return -EPERM;
	cap->free_slot = kmalloc(cap->buffer_size,GFP_KERNEL);
	if(!cap->free_slot )
		goto err_nomem;
	cap->store_slot = kmalloc(cap->buffer_size,GFP_KERNEL);
	if(!cap->store_slot)
		goto err_nomem1;
	cap->hold_slot = NULL;
	cap->store_len = 0;
	cap->hold_len = 0;
	return 0;

/* Alloc error, free allocated buffer and set proper values for structure */	
err_nomem1:
	kfree(cap->free_slot);
err_nomem:
	cap->store_len = 0;
	cap->hold_len = 0;
	cap->store_slot = NULL;
	cap->hold_slot = NULL;
	cap->free_slot = NULL;
	return -ENOMEM;
}

static void cap_free_buffer(struct capture_struct *cap)
{
	if(cap->store_slot)
		kfree(cap->store_slot);
	if(cap->hold_slot)
		kfree(cap->hold_slot);
	if(cap->free_slot)
		kfree(cap->free_slot);
	cap->store_len = 0;
	cap->hold_len = 0;
	cap->store_slot = NULL;
	cap->hold_slot = NULL;
	cap->free_slot = NULL;
}

static void __inline__ rotate_buffers( register struct capture_struct *cap)
{
	cap->hold_slot = cap->store_slot;
	cap->store_slot = cap->free_slot;
	cap->hold_len = cap->store_len;
	cap->store_len = 0;
	cap->free_slot = NULL;
}

void cap_read_timeout (unsigned long ptr)
{
	int    flags;
	struct sock * sk = (struct sock*) ptr;
	struct capture_struct * cap = &sk->protinfo.af_packet->cap;
	
	if( cap -> hold_slot == NULL ) {
		spin_lock_irqsave(&cap->lock,flags);
		if(cap->hold_slot==NULL) 
			rotate_buffers(cap);
		spin_unlock_irqrestore(&cap->lock,flags);
	}
	wake_up_interruptible(sk->sleep);
}

static int capture_rcv(struct sk_buff *skb, struct net_device *dev, 
	       	struct cap_packet_type *pt)
{
	int    flags;
	struct sock *sk = (struct sock*) pt->data;
	register struct capture_struct *cap = &sk->protinfo.af_packet->cap;
	
	atomic_inc(&cap->received);
	spin_lock_irqsave(&cap->lock,flags);
	if(cap->store_len > cap->save_store_size){
		if(cap->free_slot == NULL) {
			atomic_inc(&cap->dropped);
			goto out;
		}
		rotate_buffers(cap);
		spin_unlock_irqrestore(&cap->lock,flags);
		wake_up_interruptible(sk->sleep);
		spin_lock_irqsave(&cap->lock,flags);
	}
	if(skb->len > 0){
		cap->store_len += (*cap->type->copy)
			(cap->store_slot+cap->store_len,skb,cap->type_private);
	}
out:
	spin_unlock_irqrestore(&cap->lock,flags);
	kfree_skb(skb);
	return 0;
}

static int capture_ioctl(struct socket *sock,unsigned int cmd,unsigned long arg)
{
	struct sock *sk = sock->sk;
	struct capture_struct *cap = 
		&((struct packet_opt *)sk->protinfo.af_packet)->cap;
	int err=0;
	
	switch(cmd) 
	{
		default:
			if(cap->type->ioctl)
				err = (*cap->type->ioctl)(cmd,arg,cap->type_private);
				
	}
	return err;
}

static int capture_recvmsg(struct socket *sock, struct msghdr *msg, int len,
			  int flags, struct scm_cookie *scm)
{
	int    saveflags,err;
	struct sock *sk = sock->sk;
	struct packet_opt *opt = (struct packet_opt *)sk->protinfo.af_packet;
	struct capture_struct *cap = &opt->cap;

	msg->msg_namelen = 0;
	if (len==0) return 0;
	if (cap->buffer_size!=len) {
		if(len <= cap_copy_types[sk->type]->min_copy_size || len > CAP_MAX_BUFFERSIZE)  {
			return -EPERM;
		}
		spin_lock_irqsave(&cap->lock,saveflags);
		cap_free_buffer(cap);
		cap->buffer_size=len;
		cap->save_store_size = len - cap_copy_types[sk->type]->min_copy_size;
		err=cap_alloc_buffer(cap);
		spin_unlock_irqrestore(&cap->lock,saveflags);
		if(err)	
			return err;

		if(!sk->protinfo.af_packet->running){
			cap_add_pack(&sk->protinfo.af_packet->prot_hook);
			sk->protinfo.af_packet->running++;
		}
	}
	if (cap->read_timeout){
		cap->read_timer.expires = jiffies + cap->read_timeout*HZ/1000;
		add_timer(&cap->read_timer);
	}
	spin_lock_irqsave(&cap->lock,saveflags);
	while(cap->hold_slot == NULL) {
		if(cap->hold_slot == NULL) {
			spin_unlock_irqrestore(&cap->lock,saveflags);
			interruptible_sleep_on(sk->sleep);
			spin_lock_irqsave(&cap->lock,saveflags);
		}
		if(signal_pending(current)){
			err = -ERESTART;
			goto error_out;
		}
	}
	err = memcpy_toiovec(msg->msg_iov, cap->hold_slot, cap->hold_len);
	if(err)
		goto error_out;
	err = cap->hold_len;
	cap->hold_len=0;
	cap->free_slot=cap->hold_slot;
	cap->hold_slot=NULL;
error_out:
	spin_unlock_irqrestore(&cap->lock,saveflags);
	if(cap->read_timeout)
		del_timer(&cap->read_timer);
	return err;
}

static int capture_release(struct socket *sock)
{
	struct sock *sk = sock->sk;
	struct capture_struct *cap = &sk->protinfo.af_packet->cap;

	if (!sk)
		return 0;

	if (sk->protinfo.af_packet->running)
	{
		cap_remove_pack(&sk->protinfo.af_packet->prot_hook);
		sk->protinfo.af_packet->running = 0;
	}
	sk->state_change(sk);	
	sock->sk = NULL;
	sk->socket = NULL;
	sk->dead = 1;

	del_timer(&cap->read_timer);
	spin_lock(&cap->lock);
	if(cap->type->close)
		(*cap->type->close)(cap->type_private);
	cap_free_buffer(cap);
	sk_free(sk);
	MOD_DEC_USE_COUNT;
	return 0;
}

static int capture_create(struct socket *sock, int protocol)
{
	struct sock *sk;
	int err;
	struct capture_struct *cap;
	
	if (!capable(CAP_NET_RAW))
		return -EPERM;
	if ( cap_copy_types[sock->type] == NULL)
		return -ESOCKTNOSUPPORT;

	sock->state = SS_UNCONNECTED;
	MOD_INC_USE_COUNT;

	err = -ENOBUFS;
	sk = sk_alloc(PF_CAPTURE, GFP_KERNEL, 1);
	if (sk == NULL)
		goto out;

	sk->reuse = 1;
	sock->ops = &capture_ops;
	sock_init_data(sock,sk);

	sk->protinfo.af_packet = kmalloc(sizeof(struct packet_opt), GFP_KERNEL);
	if (sk->protinfo.af_packet == NULL)
		goto out_free;
	memset(sk->protinfo.af_packet, 0, sizeof(struct packet_opt));
	sk->zapped=0;
	sk->family = PF_CAPTURE;
	sk->num = protocol;
	cap = &sk->protinfo.af_packet->cap;
	cap->type = cap_copy_types[sk->type];

	atomic_set(&cap->received,0);
	atomic_set(&cap->dropped,0);
	spin_lock_init(&cap->lock);	
	init_timer(&cap->read_timer);
	cap->read_timer.function = cap_read_timeout;
	cap->read_timer.data = (unsigned long) sk;
	cap->read_timeout = DEF_TIMEOUT;
	if(cap_copy_types[sk->type]->open){
		err=(*cap_copy_types[sk->type]->open)
			((struct packet_opt *)sk->protinfo.af_packet);
		if(err)	
			goto out_free1;
	}

	sk->protinfo.af_packet->prot_hook.func = capture_rcv;
	sk->protinfo.af_packet->prot_hook.data = (void *)sk;

	if (protocol) {
		sk->protinfo.af_packet->prot_hook.type = protocol;
	}

	return(0);
out_free1:
	kfree(sk->protinfo.af_packet);
out_free:
	sk_free(sk);
out:
	MOD_DEC_USE_COUNT;
	return err;
}

struct proto_ops capture_ops = {
family:		PF_CAPTURE,
release:	capture_release,
bind:		sock_no_bind,
connect:	sock_no_connect,
socketpair:	sock_no_socketpair,
accept:		sock_no_accept,
getname:	sock_no_getname,
poll:		sock_no_poll,
ioctl:		capture_ioctl,
listen:		sock_no_listen,
shutdown:	sock_no_shutdown,
setsockopt:	sock_no_setsockopt,
getsockopt:	sock_no_getsockopt,
sendmsg:	sock_no_sendmsg,
recvmsg:	capture_recvmsg,
mmap:		sock_no_mmap
};

static struct net_proto_family capture_family_ops = {
	PF_CAPTURE,
	capture_create
};

int cap_copy_type_register(struct cap_copy_type *cp)
{
	if (cp->type >= MAX_CAPTYPE) {
		printk(KERN_CRIT "cap_copy_type %d >= MAX_CAP_COPY(%d)\n", cp->type, MAX_CAPTYPE);
		return -ENOBUFS;
	}
	if(cap_copy_types[cp->type])
	       return -ENOBUFS;
	cap_copy_types[cp->type]=cp;
	MOD_INC_USE_COUNT;
	return 0;
}

int cap_copy_type_unregister(unsigned short type)
{
	if (type >= MAX_CAPTYPE) {
		printk(KERN_CRIT "cap_copy_type %d >= MAX_CAPTYPE(%d)\n", type, MAX_CAPTYPE);
		return -ENOBUFS;
	}
	cap_copy_types[type]=NULL;
	MOD_DEC_USE_COUNT;
	return 0;
}

EXPORT_SYMBOL(cap_copy_type_register);
EXPORT_SYMBOL(cap_copy_type_unregister);

/*struct notifier_block packet_netdev_notifier={
	packet_notifier,
	NULL,
	0
};
*/

void cleanup_af_capture(void)
{
//	unregister_netdevice_notifier(&packet_netdev_notifier);

	sock_unregister(PF_CAPTURE);
	return;
}


int init_af_capture(void)
{
	memset(cap_copy_types,0,sizeof cap_copy_types);
	sock_register(&capture_family_ops);
	return 0;
}

module_init(init_af_capture);
module_exit(cleanup_af_capture);

