#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif
#include <linux/config.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/config.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <linux/if_ether.h>
#include <linux/ip.h>
#include "cap_pack.h"
#include "cap_lock.h"

EXPORT_SYMBOL(cap_add_pack);
EXPORT_SYMBOL(cap_remove_pack);
EXPORT_SYMBOL(netif_rx);

#if defined(MODULE_LICENSE)
        MODULE_LICENSE("GPL");
#endif


static void	sample_queue(unsigned long dummy);

int	cap_max_backlog=500;
int	cap_no_cong=40;
int	cap_lo_cong=200;
int	cap_mod_cong=480;
struct softnet_data    	cap_data[NR_CPUS] __cacheline_aligned;
static struct cap_packet_type *cap_base[16]; 
static struct cap_packet_type *cap_all = NULL;        
static struct timer_list   	cap_samp_timer = { function : sample_queue };
static struct tasklet_struct	cap_tasklet[NR_CPUS];

void cap_add_pack(struct cap_packet_type *pt)
{
        int hash;
	struct cap_packet_type **insert;
	
	if(!pt)	
		return;
	cap_write_lock_bh(CAP_CAP_LOCK);
#ifdef CONFIG_NET_FASTROUTE
        if (pt->data) {
                dev_clear_fastroute(pt->dev);
        }
#endif
        if (pt->type == htons(ETH_P_ALL)) {
		if(cap_all == NULL) {
                	pt->next=cap_all;
                	cap_all=pt;
		} else if( cap_all->prio < pt->prio) {
                	pt->next=cap_all;
                	cap_all=pt;
		} else {
			for(insert = &(cap_all->next); 
				*insert && (*insert)->prio > pt->prio; 
				insert = &((*insert)->next)) ;
			pt->next = *insert;
			*insert = pt;
		}	
        } else {
                hash=ntohs(pt->type)&15;
		if(cap_base[hash] == NULL ) {
                	pt->next = cap_base[hash];
               		cap_base[hash] = pt;
		} else if( cap_base[hash]->prio < pt->prio) {
                	pt->next = cap_base[hash];
               		cap_base[hash] = pt;
		} else {
			for(insert = &(cap_base[hash]->next); 
				*insert && (*insert)->prio > pt->prio; 
				insert = &((*insert)->next)) ;
			pt->next = *insert;
			*insert = pt;
		}
 	}
	MOD_INC_USE_COUNT;
	cap_write_unlock_bh(CAP_CAP_LOCK);
}
							      
void cap_remove_pack(struct cap_packet_type *pt)
{
        struct cap_packet_type **pt1;

        cap_write_lock_bh(CAP_CAP_LOCK);
	if (pt->type == htons(ETH_P_ALL)) {
		pt1=&cap_all;
	} else {
                pt1=&cap_base[ntohs(pt->type)&15];
        }
        for (; (*pt1) != NULL; pt1 = &((*pt1)->next)) {
                if (pt == (*pt1)) {
                      	*pt1 = pt->next;
			MOD_DEC_USE_COUNT;
			cap_write_unlock_bh(CAP_CAP_LOCK);
                        return;
                }
        }
        cap_write_unlock_bh(CAP_CAP_LOCK);
        printk(KERN_WARNING "cap_remove_pack: %p not found.\n", pt);
}

int netif_rx(struct sk_buff *skb)
{
        int this_cpu = smp_processor_id();
        struct softnet_data *queue;
        unsigned long flags;

        if (skb->stamp.tv_sec == 0)
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,18)
                get_fast_time(&skb->stamp);
#else
                do_gettimeofday(&skb->stamp);
#endif
        queue = &cap_data[this_cpu];
        local_irq_save(flags);
        if (queue->input_pkt_queue.qlen <= cap_max_backlog) {
                if (queue->input_pkt_queue.qlen) {
                        if (queue->throttle)
                                goto drop;
enqueue:
                	dev_hold(skb->dev);
			__skb_queue_tail(&queue->input_pkt_queue,skb);
#ifdef CAP_IN_KERNEL
			__cpu_raise_softirq(this_cpu, CAP_RX_SOFTIRQ);
#else
			tasklet_schedule(&cap_tasklet[this_cpu]);
#endif
			local_irq_restore(flags);
			return softnet_data[this_cpu].cng_level;
		}
                if (queue->throttle) {
		        queue->throttle = 0;
#ifdef CONFIG_NET_HW_FLOWCONTROL
		        if (atomic_dec_and_test(&netdev_dropping))
		        	netdev_wakeup();
#endif
		}
      		goto enqueue;
	}
        if (queue->throttle == 0) {
                queue->throttle = 1;
#ifdef CONFIG_NET_HW_FLOWCONTROL
                atomic_inc(&netdev_dropping);
#endif
        }

drop:
        local_irq_restore(flags);
        kfree_skb(skb);
        return NET_RX_DROP;
}

static __inline__ void skb_bond(struct sk_buff *skb)
{
	struct net_device *dev = skb->dev;

	if (dev->master) {
		dev_hold(dev->master);
		skb->dev = dev->master;
		dev_put(dev);
	}
}

#ifdef CAP_IN_KERNEL
static void cap_rx_softirq(struct softirq_action *h)
{
	int	cpu = smp_processor_id();
#else
static void cap_do_tasklet(unsigned long cpu)	
{
#endif
	struct softnet_data *queue = &cap_data[cpu];
        struct cap_packet_type *ptype;
	unsigned long start_time = jiffies;
	unsigned short type;
	int bugdet = cap_max_backlog;
	int ret = 0;

	cap_read_lock(CAP_CAP_LOCK);
	        
	for (;;) {
		struct sk_buff *skb;
		struct net_device *rx_dev;
		
		local_irq_disable();
		skb = __skb_dequeue(&queue->input_pkt_queue);
		local_irq_enable();
		
		if (skb == NULL) 	break;
		ret = 0;
		skb_bond(skb);
		rx_dev = skb->dev;
                skb->h.raw = skb->nh.raw = skb->data;
                type = skb->protocol;

		for (ptype = cap_all; ptype && ret != CAP_RX_FIN; 
					ptype = ptype->next) {
			if (!ptype->dev || ptype->dev == skb->dev) {
				atomic_inc(&skb->users);
				ret=ptype->func(skb, skb->dev, ptype);
			}
		}
		                        
		for (ptype=cap_base[ntohs(type)&15]; ptype && ret !=CAP_RX_FIN;
			       			ptype=ptype->next) {
			if (ptype->type == type && 
					(!ptype->dev||ptype->dev == skb->dev)) {
				atomic_inc(&skb->users);
				ret=ptype->func(skb, skb->dev, ptype);
			}
		}

		kfree_skb(skb);
		dev_put(rx_dev);
		
		if (bugdet-- < 0 || jiffies - start_time > 1)
			goto softnet_break;
		
#ifdef CONFIG_NET_HW_FLOWCONTROL
        	if (queue->throttle && 
				queue->input_pkt_queue.qlen < no_cong_thresh ) {
                	if (atomic_dec_and_test(&netdev_dropping)) {
	                        queue->throttle = 0;
	                        netdev_wakeup();
	                        goto softnet_break;
		        }
        	}
#endif
        }
        cap_read_unlock(CAP_CAP_LOCK);

        local_irq_disable();
        if (queue->throttle) {
                queue->throttle = 0;
#ifdef CONFIG_NET_HW_FLOWCONTROL
                if (atomic_dec_and_test(&netdev_dropping))
                        netdev_wakeup();
#endif
        }
        local_irq_enable();

        return;
softnet_break:
        cap_read_unlock(CAP_CAP_LOCK);

#ifdef CAP_IN_KERNEL	
        local_irq_disable();
        __cpu_raise_softirq(cpu, CAP_RX_SOFTIRQ);
        local_irq_enable();
#else
	tasklet_schedule(&cap_tasklet[cpu]);
#endif

        return;
}

	
static void get_sample_stats(int cpu)
{
        int blog = cap_data[cpu].input_pkt_queue.qlen;
        int avg_blog = cap_data[cpu].avg_blog;

	avg_blog = (avg_blog >> 1)+ (blog >> 1);
        if (avg_blog > cap_mod_cong) {
		cap_data[cpu].cng_level = NET_RX_CN_HIGH;
	} else if (avg_blog > cap_lo_cong) {
		cap_data[cpu].cng_level = NET_RX_CN_MOD;
	} else if (avg_blog > cap_no_cong)
		cap_data[cpu].cng_level = NET_RX_CN_LOW;
	else  
	       	cap_data[cpu].cng_level = NET_RX_SUCCESS; 
	cap_data[cpu].avg_blog = avg_blog;
}

static void sample_queue(unsigned long dummy)
{
	int	cpu=smp_processor_id();

	get_sample_stats(cpu);
	cap_samp_timer.expires = jiffies + HZ;
	add_timer(&cap_samp_timer);
}

int __init cap_init_module(void)
{
	int	i,cpu;
	struct softnet_data *queue;

	cap_all = NULL;	
	for(i = 0; i < 16; i++)
		cap_base[i] = NULL;
	
	for(cpu=0; cpu < NR_CPUS; cpu++)
	{
		queue = &cap_data[cpu];
		queue->throttle = 0;
		queue->cng_level = 0;
		queue->avg_blog = 10;
		skb_queue_head_init(&queue->input_pkt_queue);
		queue->output_queue = NULL;
		queue->completion_queue = NULL;
	}
	
	init_timer(&cap_samp_timer);
	cap_samp_timer.expires = jiffies + (1 * HZ);
	add_timer(&cap_samp_timer);
#ifdef CAP_IN_KERNEL	
	open_softirq(CAP_RX_SOFTIRQ,cap_rx_softirq,NULL);
#endif
	for(cpu=0; cpu < NR_CPUS; cpu++)
		tasklet_init(&cap_tasklet[cpu],cap_do_tasklet,cpu);
	return 0;
}

void __exit cap_exit_module(void)
{	
	int	cpu;
	
	del_timer_sync(&cap_samp_timer);
	for(cpu=0; cpu < NR_CPUS; cpu++) {
		tasklet_disable(&cap_tasklet[cpu]);
	}
}

module_init(cap_init_module);
module_exit(cap_exit_module);

/* Caplock is a copy of br_lock(Big Reader Lock).
 * When a lock is mostly lock by a reader, use caplock can speed read_lock,
 * but write lock will become slower.
 */
  
EXPORT_SYMBOL(register_caplock);
EXPORT_SYMBOL(unregister_caplock);
EXPORT_SYMBOL(__caplock_array);
EXPORT_SYMBOL(__cap_write_lock);
EXPORT_SYMBOL(__cap_write_unlock);
#ifndef __CAPLOCK_USE_ATOMICS
EXPORT_SYMBOL(__cap_write_locks);
#endif

int register_caplock(unsigned long data)
{
	MOD_INC_USE_COUNT;
	return 0;
}

int unregister_caplock(unsigned long data)
{
	MOD_DEC_USE_COUNT;
	return 0;
}

#ifdef __CAPLOCK_USE_ATOMICS

caplock_read_lock_t __caplock_array[NR_CPUS][__CAP_IDX_MAX] =
   { [0 ... NR_CPUS-1] = { [0 ... __CAP_IDX_MAX-1] = RW_LOCK_UNLOCKED } };

void __cap_write_lock (enum caplock_indices idx)
{
	int i;

	for (i = 0; i < smp_num_cpus; i++)
		write_lock(__caplock_array[idx] + cpu_logical_map(i));
}

void __cap_write_unlock (enum caplock_indices idx)
{
	int i;

	for (i = 0; i < smp_num_cpus; i++)
		write_unlock(__caplock_array[idx] + cpu_logical_map(i));
}

#else 

caplock_read_lock_t __caplock_array[NR_CPUS][__CAP_IDX_MAX] =
   { [0 ... NR_CPUS-1] = { [0 ... __CAP_IDX_MAX-1] = 0 } };

struct cap_wrlock __cap_write_locks[__CAP_IDX_MAX] =
   { [0 ... __CAP_IDX_MAX-1] = { SPIN_LOCK_UNLOCKED } };

void __cap_write_lock (enum caplock_indices idx)
{
	int i;

	spin_lock(&__cap_write_locks[idx].lock);
again:
	for (i = 0; i < smp_num_cpus; i++)
		if (__caplock_array[cpu_logical_map(i)][idx] != 0)
			goto again;
}

void __cap_write_unlock (enum caplock_indices idx)
{
	spin_unlock(&__cap_write_locks[idx].lock);
}

#endif 
