#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif
#include <linux/config.h>
#include <linux/types.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/ip.h>
#include <linux/in.h>
#include <linux/tcp.h>
#include <linux/udp.h>
#include <net/checksum.h>
#include "af_capture.h" 
#include "cap_type.h"

#if defined(MODULE_LICENSE)
        MODULE_LICENSE("GPL");
#endif

static int cap_copy_hdr_open(struct packet_opt *opt)
{
	MOD_INC_USE_COUNT;
	opt->prot_hook.prio=CAP_PRIO_COPY_HDR;
	return 0;
}
static void cap_copy_hdr_close(void *priviate_data)
{
	MOD_DEC_USE_COUNT;
}
static int cap_copy_hdr(char *dst, register struct sk_buff *skb, void *private)
{
	struct iphdr *iph= skb->nh.iph;
	register struct hdr_type *store=(struct hdr_type*) dst;

	if (skb->len < sizeof(struct iphdr))
		goto error; 
	if (iph->ihl < 5 || iph->version != 4 || 
			ip_fast_csum((u8 *)iph, iph->ihl) != 0)
		goto error; 

	store->stamp=skb->stamp;
	memcpy(store->data,skb->h.raw,HDRLEN);
	
	return sizeof (struct hdr_type);	
error:
	return 0;
}
static struct cap_copy_type cap_type_hdr = {
	CAP_COPY_HDR,
	sizeof(struct hdr_type),
	cap_copy_hdr,
	cap_copy_hdr_open,
	cap_copy_hdr_close,
	NULL
};


static int cap_copy_flow_open(struct packet_opt *opt)
{
	MOD_INC_USE_COUNT;
	opt->prot_hook.prio=CAP_PRIO_COPY_FLOW;
	return 0;
}

static void cap_copy_flow_close(void *private_data)
{
	MOD_DEC_USE_COUNT;
}

static int cap_copy_flow(char *dst, register struct sk_buff *skb, void *private)
{
	struct iphdr *iph = skb->nh.iph;
	register struct flow_type *store=(struct flow_type*) dst;
	int len;

	if (skb->len < sizeof(struct iphdr))
		goto error; 
	if (iph->ihl < 5 || iph->version != 4 || 
			ip_fast_csum((u8 *)iph, iph->ihl) != 0)
		goto error; 

	len = ntohs(iph->tot_len); 
	if (skb->len < len)
		goto error; 
	__skb_trim(skb, len);
	len=skb->len;
//
//	if the ip header is not counted:
//	len -= sizeof(struct iphdr);
//
	if (len < 0)
		goto error;
	store->length = len;
	store->protocol = iph->protocol;
	store->saddr = iph->saddr;
	store->daddr = iph->daddr;
	store->time = skb->stamp.tv_sec;
	skb->h.raw = skb->nh.raw + iph->ihl*4;
	switch (iph->protocol){
	case IPPROTO_TCP:
		if(skb->len >= sizeof (struct tcphdr)){
			store->sport=ntohs(skb->h.th->source);
			store->dport=ntohs(skb->h.th->dest);
			goto out; 
		}
		store->sport = 0;
		store->dport = 0;
		goto out;

	case IPPROTO_UDP:
		if(skb->len >= sizeof (struct udphdr)){
			store->sport=ntohs(skb->h.uh->source);
			store->dport=ntohs(skb->h.uh->dest);
			goto out;
		}
	default:
		store->sport =0;
		store->dport =0;
		goto out;
	}
out:
	return sizeof (struct flow_type);	
error:
	return 0;
}

static struct cap_copy_type cap_type_flow = {
	CAP_COPY_FLOW,
	sizeof(struct flow_type),
	cap_copy_flow,
	cap_copy_flow_open,
	cap_copy_flow_close,
	NULL
};

static int cap_copy_var_open(struct packet_opt *opt)
{
	MOD_INC_USE_COUNT;
	opt->prot_hook.prio=CAP_PRIO_COPY_VAR;
	return 0;
}

static void cap_copy_var_close(void *private_data)
{
	MOD_DEC_USE_COUNT;
}

static int cap_copy_var(char *dst, register struct sk_buff *skb, void *private)
{
	struct iphdr *iph = skb->nh.iph;
	register struct var_type *store=(struct var_type*) dst;
	int len;

	if (skb->len < sizeof(struct iphdr))
		goto error; 
	if (iph->ihl < 5 || iph->version != 4 || 
			ip_fast_csum((u8 *)iph, iph->ihl) != 0)
		goto error; 

	len = ntohs(iph->tot_len); 
	if (skb->len < len)
		goto error; 
	__skb_trim(skb, len);
	len=skb->len;
	len -= sizeof(struct iphdr);
	if (len < 0)
		goto error;
#ifdef COPY_VAR_IP_VERSION
	store->version = iph->version | iph->ihl;
#endif
#ifdef COPY_VAR_IP_TOS
	store->tos = iph->tos;
#endif
#ifdef COPY_VAR_IP_TOTAL_LEN
	store->total_len = iph->tot_len;
#endif
#ifdef COPY_VAR_IP_ID
	store->id = iph->id;
#endif
#ifdef COPY_VAR_IP_FRAGOFF
	store->fragoff = iph->frag_off;
#endif
#ifdef COPY_VAR_IP_TTL
	store->ttl = iph->ttl;
#endif
#ifdef COPY_VAR_IP_PROTOCOL
	store->protocol = iph->protocol;
#endif
#ifdef COPY_VAR_IP_CHECK
	store->check = iph->check;
#endif
#ifdef COPY_VAR_IP_SADDR
	store->saddr = iph->saddr;
#endif
#ifdef COPY_VAR_IP_DADDR
	store->daddr = iph->daddr;
#endif
//	store->time = skb->stamp.tv_sec;
	skb->h.raw = skb->nh.raw + iph->ihl*4;
	switch (iph->protocol){
	case IPPROTO_TCP:
		if(skb->len >= sizeof (struct tcphdr)){
#ifdef COPY_VAR_TCP_SPORT			
			store->proto.tcphdr.sport=ntohs(skb->h.th->source);
#endif
#ifdef COPY_VAR_TCP_DPORT			
			store->proto.tcphdr.dport=ntohs(skb->h.th->dest);
#endif
			break;
		}
#ifdef COPY_VAR_TCP_SPORT			
		store->proto.tcphdr.sport = 0;
#endif		
#ifdef COPY_VAR_TCP_DPORT			
		store->proto.tcphdr.dport = 0;
#endif
		break;	

	case IPPROTO_UDP:
		if(skb->len >= sizeof (struct udphdr)){
#ifdef COPY_VAR_UDP_SPORT			
			store->proto.udphdr.sport=ntohs(skb->h.uh->source);
#endif
#ifdef COPY_VAR_UDP_DPORT			
			store->proto.udphdr.dport=ntohs(skb->h.uh->dest);
#endif
			break;
		}
#ifdef COPY_VAR_UDP_SPORT			
		store->proto.udphdr.sport = 0;
#endif
#ifdef COPY_VAR_UDP_DPORT			
		store->proto.udphdr.dport = 0;
#endif

	default:
//		store->sport =0;
//		store->dport =0;
		break;
	}
        return sizeof (struct var_type);
error:
        return 0;
}	

static struct cap_copy_type cap_type_var = {
	CAP_COPY_VAR,
	sizeof(struct var_type),
	cap_copy_var,
	cap_copy_var_open,
	cap_copy_var_close,
	NULL
};

int __init cap_type_init_module(void)
{
	cap_copy_type_register(&cap_type_hdr);
	cap_copy_type_register(&cap_type_flow);
	cap_copy_type_register(&cap_type_var);
	return 0;
}

void __exit cap_type_exit_module(void)
{
	cap_copy_type_unregister(CAP_COPY_HDR);
	cap_copy_type_unregister(CAP_COPY_FLOW);
	cap_copy_type_unregister(CAP_COPY_VAR);
}

module_init(cap_type_init_module);
module_exit(cap_type_exit_module);
