/*
   capdump.c
   A demo program use the AF_CAPTURE socket interface to capture the packet.
   This program will print out the IP protocol packet head it knows.
   You can use:
   capdump | egrep -v "TCP|UDP|ICMP" 
   to see what protocol left in the your ethernet. I can found IPV6 turnnel
   here.
   
NOTICE:
   Read capture.txt first to install the capture modules.
   "ifconfig ethX promisc" to set your NIC in promisc mode by hand. This
   program will not set it for you.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <asm/types.h>
#include <linux/if_ether.h>
#include <linux/in.h>
#include "cap_pack.h"
#include "cap_type.h"
#include "af_capture.h"
#include "headfields.h"
			

/* I have noway to use inet_ntoa() without copy the ip address to 

   a struct in_addr. Any way, write one is not complicate */
   
void printf_ip(unsigned long ip)
{
	printf("%ld.%ld.%ld.%ld",ip&0xff,(ip>>8)&0xff,(ip>>16)&0xff,(ip>>24)&0xff);
}

int main(void)
{
	int sk,i;
	struct hdr_type *hdr;
	ssize_t count;
	char buffer[64000];
	sk = socket( AF_CAPTURE, CAP_COPY_HDR,ntohs(ETH_P_IP));
	//for(i=0;i<100;i++){
	while(1) {
		count = read(sk,buffer,64000);
		for(hdr=(struct hdr_type*) buffer;(char*)hdr<buffer+count;hdr++){
		if (IPHDRLEN(hdr)>20) {	
			printf("IPHDR:%d: ",IPHDRLEN(hdr));
			printf_ip(IPSADDR(hdr));
			printf(" => ");
			printf_ip(IPDADDR(hdr));
			printf(" ");
			switch(IPPROTL(hdr)){
			case IPPROTO_IP:
				printf("DUMMY");
				break;
			case IPPROTO_TCP:
				printf("TCP ");
				printf("port %d => %d",ntohs(TCPSPORT(hdr)),ntohs(TCPDPORT(hdr)));
				break;
			case IPPROTO_UDP:
				printf("UDP ");
				printf("port %d => %d",ntohs(UDPSPORT(hdr)),ntohs(UDPDPORT(hdr)));
				break;
			case IPPROTO_IDP:
				printf("XNS IDP");
				break;
			case IPPROTO_PUP:
				printf("PUP");
				break;
			case IPPROTO_GRE:
				printf("Cisco GRE");
				break;
			case IPPROTO_ICMP:
				printf("ICMP");
				break;
			case IPPROTO_IGMP:
				printf("IGMP");
				break;
			case IPPROTO_IPIP:
				printf("IPIP tunnel");
				break;
			case IPPROTO_EGP:
				printf("EGP Exterior Gateway Protocol");
				break;
			case IPPROTO_IPV6:
				printf("IPV6 in IPV4 tunnelling");
				break;
			case IPPROTO_RSVP:
				printf("RSVP");
				break;
			case IPPROTO_PIM:
				printf("PIM");
				break;
			default:
				printf("Unkown protocol ID %d",IPPROTL(hdr));
			}
			printf(" size = %d ttl=%d\n",ntohs(IPLEN(hdr)),IPTTL(hdr));
		}	
		}
	}
	return 0;
}

	

