/*
 * 	LPF:	IP packet header analysis fields selection
 *
 * 	Authors:
 * 		ZhiChun Li<lizc@ieee.org>
 * 	Descrition:
 * 	  LPF has defined CAP_COPY_HEADER kernel filter to get pakcet time stamp
 * 	  and first 40 bytes of IP packet. 
 * 	  In this file we has define some macro to select the IP header or TCP (UDP etc) header
 * 	  fields for post-analysis.
 */

#include "cap_type.h"
/* 
 * These micros are defined for selecting IP &TCP(UDP,ICMP,IGMP) header field
 * can add more for more fields
 */
#define IPHDRLEN(x) ((((struct hdr_record *)(x))->iph.ihl)*4)      /* IP header length */ 
#define IPSADDR(x) (((struct hdr_record *)(x))->iph.saddr)  	/* 32bits IP Address */
#define IPDADDR(x) (((struct hdr_record *)(x))->iph.daddr) 	/* 32bits IP Address */
#define IPPROTL(x) (((struct hdr_record *)(x))->iph.protocol)	/* 8bits IP protocol type */
#define IPLEN(x)   (((struct hdr_record *)(x))->iph.tot_len)	/* 16bits total IP length */
#define IPTTL(x)   (((struct hdr_record *)(x))->iph.ttl) 	/* 8bits TTL */
#define IPTOS(x)   (((struct hdr_record *)(x))->iph.tos)        /* 8bits TOS */	
#define IPFRAG(x)   (((struct hdr_record *)(x))->iph.frag_off)        /* 16bits FRAGMENT */

#if defined(__LITTLE_ENDIAN_BITFIELD)
#define IPFRAG_RF(x) (IPFRAG(x) & (0x8000))
#define IPFRAG_DF(x) (IPFRAG(x) & (0x4000))
#define IPFRAG_MF(x) (IPFRAG(x) & (0x2000))
#define IPFRAG_OFFMASK(x) (IPFRAG(x) & (0x1FFF))
#elif defined (__BIG_ENDIAN_BITFIELD)
#define IPFRAG_RF(x) (IPFRAG(x) & (0x0001))
#define IPFRAG_DF(x) (IPFRAG(x) & (0x0002))
#define IPFRAG_MF(x) (IPFRAG(x) & (0x0004))
#define IPFRAG_OFFMASK(x) ((IPFRAG(x) & (0xFFF8))/8)
#else
#error  "Please fix <asm/byteorder.h>"
#endif
/*
 * Attention please, IF IPHDRLEN>20, which means the ip header used ip options, cannot use the type of macros
 * must use the TRANSPORTHDR macro to get the beginning of the transport layer header position
 * Why define this type of macro? for easy use. Because the IP packet which has ip options is too few in the
 * real network traffic about 1/10000 less.
 */
#define TCPSPORT(x) 	(((struct hdr_record *)(x))->h.tcph.source) 	/* 16bits Port */
#define TCPDPORT(x) 	(((struct hdr_record *)(x))->h.tcph.dest)	/* 16bits Port */
#define TCPSEQ(x)	(((struct hdr_record *)(x))->h.tcph.seq)	/* 32bits seq number */
#define TCPACK_SEQ(x)	(((struct hdr_record *)(x))->h.tcph.ack_seq)	/* 32bits ack_seq number */
#define TCPRES1(x)	(((struct hdr_record *)(x))->h.tcph.res1)	/* 4bits res1 */
#define TCPDOFF(x)	(((struct hdr_record *)(x))->h.tcph.doff)	/* 4bits doff */
#define TCPFIN(x)	(((struct hdr_record *)(x))->h.tcph.fin)	/* 1bits fin flag */
#define TCPSYN(x)	(((struct hdr_record *)(x))->h.tcph.syn)	/* 1bits syn flag */
#define TCPRST(x)	(((struct hdr_record *)(x))->h.tcph.rst)	/* 1bits rst flag */
#define TCPPSH(x)	(((struct hdr_record *)(x))->h.tcph.psh)	/* 1bits psh flag */
#define TCPACK(x)	(((struct hdr_record *)(x))->h.tcph.ack)	/* 1bits ack flag */
#define TCPURG(x)	(((struct hdr_record *)(x))->h.tcph.urg)	/* 1bits urg flag */
#define TCPECE(x)	(((struct hdr_record *)(x))->h.tcph.ece)	/* 1bits ece flag */
#define TCPCWR(x)	(((struct hdr_record *)(x))->h.tcph.cwr)	/* 1bits cwr flag */
#define TCPWIN(x)	(((struct hdr_record *)(x))->h.tcph.window)	/* 16bits window flag */

#define UDPSPORT(x) 	(((struct hdr_record *)(x))->h.udph.source)	/* 16bits Port */
#define UDPDPORT(x) 	(((struct hdr_record *)(x))->h.udph.dest)	/* 16bits Port */
#define UDPLEN(x)	(((struct hdr_record *)(x))->h.udph.len)	/* 16bits Len  */

#define ICMPTYPE(x)	(((struct hdr_record *)(x))->h.icmph.type)	/* 8bits icmp message type */
#define ICMPCODE(x)	(((struct hdr_record *)(x))->h.icmph.code)	/* 8bits icmp message code */

#define IGMPTYPE(x)	(((struct hdr_record *)(x))->h.igmph.type)	/* 8bits igmp message type */
#define IGMPCODE(x)	(((struct hdr_record *)(x))->h.igmph.code)	/* 8bits igmp message code */
#define IGMPGROUP(x)	(((struct hdr_record *)(x))->h.igmph.group)	/* 16bits igmp message group */

#define TRANSPORTHDR(x) ((void *)(x+sizeof(struct timeval)+IPHDRLEN(x))) /*return a pointer point to
									   the beginning of transport 
									   layer header; use in the case 
									   which IPHDRLEN>20, the previous
									   hdr_record struct cannot be fitted.
									  */
/*
 * These two function are used to get PORT no matter how the the transport protocol are TCP or UDP
 * if return port=0 may means not TCP or UDP 
 */
unsigned short IPSPORT (struct hdr_type* hdr)
{
	unsigned short register port;
	switch (IPPROTL(hdr)) {
	case IPPROTO_TCP:
		if (ntohs(IPLEN(hdr)) >= sizeof(struct tcphdr) + sizeof(struct iphdr)) {
			port=TCPSPORT(hdr);
			return port;
		}
		port=0;
		return port;
	case IPPROTO_UDP:
		if (ntohs(IPLEN(hdr)) >= sizeof(struct udphdr) + sizeof(struct iphdr)) {
			port=UDPSPORT(hdr);
			return port;
		}
		port=0;
		return port;
	default:
		port=0;
		return port;
	}
	return port;
}
unsigned short IPDPORT (struct hdr_type* hdr)
{
	unsigned short register port;
	switch (IPPROTL(hdr)) {
	case IPPROTO_TCP:
		if (ntohs(IPLEN(hdr)) >= sizeof(struct tcphdr) + sizeof(struct iphdr)) {
			port=TCPDPORT(hdr);
			return port;
		}
		port=0;
		return port;
	case IPPROTO_UDP:
		if (ntohs(IPLEN(hdr)) >= sizeof(struct udphdr) + sizeof(struct iphdr)) {
			port=UDPDPORT(hdr);
			return port;
		}
		port=0;
		return port;
	default:
		port=0;
		return port;
	}
	return port;
}


	

